# Copyright 2012, 2013 Marco Ceppi, Canonical Ltd.  This software is
# licensed under the GNU Affero General Public License version 3 (see
# the file LICENSE).

"""
Charm review queue

 Contributed by Juan Negronil


 Changes
 - convert to job.
 - use dateutil date parsing instead of string comparsion.
 - merge proposals show target with series/branch.

"""
import datetime
import itertools
import logging

from launchpadlib.launchpad import Launchpad
import pymongo

from charmworld.utils import (
    configure_logging,
    get_ini,
    Timer,
)
from config import MONGO_URL
from utils import lock
from utils import LockHeld


def get_tasks_and_proposals(log):
    '''Gets review items (bugs and proposals) from launchpad.'''
    # TODO: Eventually these single requests should be broken into two each:
    # one to get anything created within the window, and one to get anything
    # older than the window that is still open. Right now this is probably the
    # more performant way to go about it, but that's going to change as the
    # data outside of our window gets larger.
    with Timer() as timer:
        lp = Launchpad.login_anonymously('charm-tools',
                                         'production',
                                         version='devel')

        # Entity roots for items of interests
        charm = lp.distributions['charms']
        charmers = lp.people['charmers']
        charm_contributors = lp.people['charm-contributors']

    log.debug("lp login time %0.2f", timer.duration())

    with Timer() as timer:
        bugs = charm.searchTasks(
            tags=['new-formula', 'new-charm'],
            tags_combinator="Any")
        # Best practice  dictates that charmers be subscribed to all bugs
        # and merge proposals for official source package branches.
        charmers_bugs = charmers.searchTasks()
        proposals = charmers.getRequestedReviews()
        contributor_proposals = charm_contributors.getRequestedReviews()

        juju = lp.projects['juju']
        doc_proposals = [mp for mp in juju.getMergeProposals()
                         if mp.target_branch_link.endswith("~juju/juju/docs")]

    log.debug("lp query time %0.2f", timer.duration())
    return (itertools.chain(bugs, charmers_bugs),
            itertools.chain(proposals, contributor_proposals, doc_proposals))


def update_review_queue(db, tasks, proposals, log):
    '''Updates the review queue information.'''
    bug_statuses = [
        'New', 'Confirmed', 'Triaged', 'In Progress', 'Fix Committed']
    proposal_statuses = ['Needs review', 'Needs fixing', 'Needs information']
    now = datetime.datetime.utcnow()
    window = now - datetime.timedelta(days=180)  # 6 month window for latency

    r_queue = []
    latencies = []
    for task in tasks:
        if (task.date_created.replace(tzinfo=None) >= window or
            task.status in bug_statuses):
            if task.date_confirmed:
                latency = task.date_confirmed - task.date_created
            else:
                # Now has been calculated in UTC, and LP returns UTC TZ
                # datetimes. We can safely remove the TZ, letting us get the
                # timedelta.
                latency = now - task.date_created.replace(tzinfo=None)
            latencies.append(latency.total_seconds())

        if task.status in bug_statuses:
            entry = {
                'type': 'task',
                'date_modified': task.bug.date_last_updated,
                'date_created': task.date_created,
                'summary': task.title.split('"')[1].strip(),
                'item': task.web_link,
                'status': task.status,
            }
            r_queue.append(entry)

    for proposal in proposals:
        if (proposal.date_created.replace(tzinfo=None) >= window or
            proposal.queue_status in proposal_statuses):
            if proposal.date_reviewed:
                latency = (
                    proposal.date_reviewed - proposal.date_review_requested)
            else:
                # Now has been calculated in UTC, and LP returns UTC TZ
                # datetimes. We can safely remove the TZ, letting us get the
                latency = (
                    now - proposal.date_review_requested.replace(tzinfo=None))
            latencies.append(latency.total_seconds())

        if proposal.queue_status in proposal_statuses:
            parts = proposal.target_branch_link.rsplit('/', 3)
            if parts[-1] == "trunk":
                parts.pop(-1)
            target = "/".join(parts[2:])
            origin = "/".join(proposal.source_branch_link.rsplit('/', 3)[1:])

            num_comments = proposal.all_comments.total_size
            if num_comments != 0:
                last_modified = proposal.all_comments[
                    num_comments - 1].date_created
            else:
                last_modified = proposal.date_created

            entry = {
                'type': 'proposal',
                'latency': latency.total_seconds(),
                'date_modified': last_modified,
                'date_created': proposal.date_created,
                'summary': "Merge for %s from %s" % (target, origin),
                'item': proposal.web_link,
                'status': proposal.queue_status,
            }
            r_queue.append(entry)

    # Drop old collection
    log.info(
        "Dropping old review queue (%d items)", db['review_queue'].count())
    db.drop_collection('review_queue')

    # Bulk insert
    log.info("Updating review queue (%d items)" % len(r_queue))
    if (r_queue != []):
        # Needs check as bulk insert of an empty list is prohibited
        db['review_queue'].insert(r_queue)

    # Get stats:
    stats = latency_stats(latencies)
    if stats:
        stats['date'] = now
        db.review_queue_latency.insert(stats)


def latency_stats(queue):
    if queue != []:
        min_latency = min(queue)
        mean_latency = sum(queue) / len(queue)
        max_latency = max(queue)
        return {'min': min_latency, 'max': max_latency, 'mean': mean_latency}


def main():
    configure_logging()
    log = logging.getLogger("charm.review")
    c = pymongo.Connection(MONGO_URL, safe=True)
    db = c['juju']
    settings = get_ini()
    try:
        with lock('ingest-review',
                  int(settings['script_lease_time']) * 60, db, log):
            tasks, proposals = get_tasks_and_proposals(log)
            update_review_queue(db, tasks, proposals, log)
    except LockHeld, error:
        log.warn(str(error))

if __name__ == '__main__':
    main()
