# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import datetime
import logging

from charmworld.jobs.review import update_review_queue
from charmworld.testing import MongoTestBase


NOW = datetime.datetime.now()
YESTERDAY = NOW - datetime.timedelta(1)
TWO_DAYS = NOW - datetime.timedelta(2)


class MockBug:
    date_last_updated = NOW


class MockTask:

    def __init__(self, confirmed=False):
        self.date_created = TWO_DAYS
        self.title = u'Bug #0 in a collection: "A bug task"'
        self.web_link = "http://example.com"
        self.bug = MockBug()
        if confirmed:
            self.status = "Confirmed"
            self.date_confirmed = YESTERDAY
        else:
            self.date_confirmed = None
            self.status = "New"


class MockCommentCollection:
    total_size = 0


class MockProposal:

    def __init__(self, reviewed=False):
        base = 'https://api.launchpad.net/devel/'
        self.target_branch_link = base + '~charmers/charms/precise/foo/trunk'
        self.source_branch_link = base + '~bar/charms/precise/foo/fnord'
        self.all_comments = MockCommentCollection()
        self.date_review_requested = TWO_DAYS
        self.date_created = TWO_DAYS
        self.web_link = 'http://example.com'
        if reviewed:
            self.date_reviewed = YESTERDAY
            self.queue_status = "Approved"
        else:
            self.date_reviewed = None
            self.queue_status = "Needs review"


class ReviewTest(MongoTestBase):

    def test_review_handles_empty_queue(self):
        self.db.review_queue.insert([{'foo': 'bar'}])
        log = logging.getLogger('foo')
        update_review_queue(self.db, [], [], log)
        self.assertEqual(0, self.db.review_queue.count())

    def test_review_handles_bugs(self):
        log = logging.getLogger('foo')
        update_review_queue(self.db, [MockTask()], [], log)
        self.assertEqual(1, self.db.review_queue.count())
        item = self.db.review_queue.find_one()
        self.assertEqual(NOW.date(), item['date_modified'].date())
        self.assertEqual(TWO_DAYS.date(), item['date_created'].date())
        self.assertEqual('A bug task', item['summary'])
        self.assertEqual('http://example.com', item['item'])
        self.assertEqual('New', item['status'])

    def test_review_handles_proposals(self):
        log = logging.getLogger('foo')
        update_review_queue(self.db, [], [MockProposal()], log)
        self.assertEqual(1, self.db.review_queue.count())
        item = self.db.review_queue.find_one()
        self.assertEqual(TWO_DAYS.date(), item['date_modified'].date())
        self.assertEqual(TWO_DAYS.date(), item['date_created'].date())
        self.assertEqual(
            'Merge for foo from precise/foo/fnord', item['summary'])
        self.assertEqual('http://example.com', item['item'])
        self.assertEqual('Needs review', item['status'])

    def test_review_calculates_latency(self):
        log = logging.getLogger('foo')
        self.assertEqual(0, self.db.review_queue_latency.count())
        update_review_queue(self.db, [], [MockProposal(True)], log)
        self.assertEqual(1, self.db.review_queue_latency.count())
        item = self.db.review_queue_latency.find_one()
        self.assertEqual(NOW.date(), item['date'].date())
        self.assertEqual(1, datetime.timedelta(seconds=item['max']).days)
        self.assertEqual(1, datetime.timedelta(seconds=item['mean']).days)
        self.assertEqual(1, datetime.timedelta(seconds=item['min']).days)

    def test_review_discards_closed_bugs(self):
        tasks = [MockTask(), MockTask()]
        tasks[0].status = 'Fix released'
        log = logging.getLogger('foo')
        update_review_queue(self.db, tasks, [], log)
        self.assertEqual(1, self.db.review_queue.count())

    def test_review_discards_approved_proposals(self):
        proposals = [MockProposal(), MockProposal()]
        proposals[0].queue_status = 'Approved'
        log = logging.getLogger('foo')
        update_review_queue(self.db, [], proposals, log)
        self.assertEqual(1, self.db.review_queue.count())
