# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import textwrap
import yaml

from charmworld.models import (
    Bundle,
    CharmSource,
    store_bundles,
)
from charmworld.search import (
    BUNDLE,
    IndexMissing,
)
from charmworld.testing import (
    factory,
    MongoTestBase,
)
from charmworld.migrations import migrate
from charmworld.migrations.migrate import Versions

import mock


class MigrationTestBase(MongoTestBase):

    def setUp(self):
        super(MigrationTestBase, self).setUp()
        self.versions = Versions(migrate.get_migration_path())
        self.use_context(mock.patch.object(migrate, 'configure_logging'))


class TestMigration018(MigrationTestBase):

    def setUp(self):
        super(TestMigration018, self).setUp()
        self.use_index_client()
        self.store_some_bundles()

    def store_some_bundles(self):
        deployer_config = textwrap.dedent("""\
            wordpress-stage:
                series: precise
                services:
                    blog:
                        charm: cs:precise/wordpress
                        constraints: mem=2
                        options:
                            tuning: optimized
                            engine: apache
        """)
        parsed = yaml.safe_load(deployer_config)
        owner = 'owner'
        bundle_name = 'wordpress-stage'
        basket_name = 'wordpress-basket'
        basket_rev = 3
        basket_id = "%s/%d" % (basket_name, basket_rev)
        self.search_id = Bundle.construct_id(
            owner, basket_name, bundle_name)
        store_bundles(
            self.db.bundles, parsed, owner, basket_id,
            index_client=self.index_client)

    def test_bundles_are_removed_from_elastic_search(self):
        assert self.index_client.get(self.search_id, BUNDLE), (
            'the test bundle is missing')
        self.versions.run_migration(
            self.db, self.index_client, '018_delete_all_bundles.py')
        self.assertIsNone(self.index_client.get(self.search_id, BUNDLE))

    def test_bundles_are_removed_from_mongo(self):
        assert self.index_client.get(self.search_id, BUNDLE), (
            'the test bundle is missing')
        self.versions.run_migration(
            self.db, self.index_client, '018_delete_all_bundles.py')
        self.assertIsNone(self.db.bundles.find_one({}))

    def test_new_db_and_index_are_noop(self):
        # The migration script survives MIssingIndex errors.
        # New installs don't have ES indexes yet.
        self.index_client.delete_index()
        with self.assertRaises(IndexMissing):
            self.index_client.get_status()
        self.versions.run_migration(
            self.db, self.index_client, '018_delete_all_bundles.py')
        # Put the index back for the test case to handle.
        self.index_client.create_index()


class TestMigration019(MigrationTestBase):

    def setUp(self):
        super(TestMigration019, self).setUp()
        self.use_index_client()

    def test_new_db_and_index_have_all_old_data(self):
        source = CharmSource.from_request(self)
        charm_1 = factory.get_charm_json(
            name='one',
            provides={
                'website': {
                    'interface': 'http',
                },
            },
            requires={
                'database': {
                    'interface': 'mysql',
                }
            })
        source.save(charm_1)
        charm_2 = factory.get_charm_json(
            name='two',
            provides={
                'database': {
                    'interface': 'mysql',
                }
            })
        source.save(charm_2)

        exodus_update = self.versions.get_exodus(
            '019_fixed_mapping_for_interfaces.py')
        target = self.versions.get_pending_source(source, 19)
        target.index_client.create_index()
        target.index_client.put_mapping()
        self.addCleanup(target.index_client.delete_index)
        exodus_update(source, target, 19)
        self.assertEqual(2, target.collection.find().count())
        self.assertEqual(2, target.index_client.search('mysql')['matches'])
        self.assertEqual(1, target.index_client.search('http')['matches'])


class TestMigration020(MigrationTestBase):

    def setUp(self):
        super(TestMigration020, self).setUp()
        self.use_index_client()

    def test_review_entries_updated(self):
        review_entries = [
            factory.make_review_entry(),
            factory.make_review_entry()]
        review_entries[0]['item'] = 'code.launchpad.net'
        review_entries[1]['item'] = 'bugs.launchpad.net'
        self.db.review_queue.insert(review_entries)
        self.versions.run_migration(
            self.db, self.index_client,
            '020_adds_type_field_to_existing_review_items.py')
        results = list(self.db.review_queue.find())
        self.assertEqual(2, len(results))
        for entry in results:
            if entry['item'] == 'code.launchpad.net':
                self.assertEqual('proposal', entry['type'])
            elif entry['item'] == 'bugs.launchpad.net':
                self.assertEqual('task', entry['type'])
            else:
                self.fail()

    def test_fails_on_unknown(self):
        review_entries = [factory.make_review_entry()]
        review_entries[0]['item'] = 'example.com'
        self.db.review_queue.insert(review_entries)
        self.assertRaises(
            Exception,
            self.versions.run_migration,
            self.db, self.index_client,
            '020_adds_type_field_to_existing_review_items.py')
