# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type


from datetime import datetime

from charmworld.models import (
    Charm,
    QADataSource,
)
from charmworld.views.tools import (
    charm_review,
    charm_store_missing,
    get_proof_errors_for_queue,
    get_sparklines,
    get_store_errors_for_queue,
    get_missing_qa_for_queue,
    proof_errors,
)
from charmworld.testing import factory
from charmworld.testing import ViewTestBase
from charmworld.testing import WebTestBase
from charmworld.testing.html import parse


class TestToolsHome(WebTestBase):

    def test_tools_view(self):
        """Make sure the tools view renders correctly."""
        resp = self.app.get(
            '/tools',
            status=200)

        body = resp.body
        self.assertIn("Charm Tools", body)


class TestReviews(WebTestBase):

    def test_last_modified_formatting(self):
        self.db.review_queue.save(
            factory.make_review_entry(date_modified=datetime(2012, 1, 2)))
        self.db.review_queue.save(
            factory.make_review_entry(date_modified=datetime(2012, 1, 3)))
        resp = self.app.get('/tools/review-queue')
        tree = parse(resp.body)
        headings = [th.text for th in tree.findall('body//table/thead/tr/th')]
        self.assertEqual(
            ['Age', 'Date Created', 'Last Modified', 'Summary', 'Status'],
            headings)
        rows = tree.findall('body//table/tr')
        self.assertEqual(2, len(rows))
        colmap = {}
        for row in rows:
            columns = row.findall('td')
            self.assertEqual(len(headings), len(columns))
            [colmap.setdefault(heading, []).append(col.text)
             for heading, col in zip(headings, columns)]
        self.assertEqual(['2012/01/02', '2012/01/03'], colmap['Last Modified'])

    def test_empty_sparklines(self):
        sparklines = get_sparklines([])
        self.assertEqual('', sparklines['maxes'])
        self.assertEqual('', sparklines['mins'])
        self.assertEqual('', sparklines['means'])

    def test_single_sparklines(self):
        sparklines = get_sparklines([{'max': 1, 'min': 2, 'mean': 3}])
        self.assertEqual('1', sparklines['maxes'])
        self.assertEqual('2', sparklines['mins'])
        self.assertEqual('3', sparklines['means'])

    def test_sparklines(self):
        sparklines = get_sparklines([
            {'max': 1, 'min': 1, 'mean': 1},
            {'max': 1, 'min': 2, 'mean': 3}])
        self.assertEqual('1,1', sparklines['maxes'])
        self.assertEqual('1,2', sparklines['mins'])
        self.assertEqual('1,3', sparklines['means'])


class TestExpandedItemsInQueue(ViewTestBase):

    def test_missing_qa_data_in_queue(self):
        factory.makeCharm(self.db, promulgated=True)
        qa = get_missing_qa_for_queue(self.getRequest())
        self.assertEqual(1, len(qa))

    def test_charm_with_qa_data_matches_name_and_series(self):
        # Only one of the two charms should make it into the queue.
        _, charm_json = factory.makeCharm(self.db, promulgated=True)
        charm = Charm(charm_json)
        factory.makeCharm(
            self.db,
            name=charm.name,
            series='a-different-series',
            promulgated=True
        )
        qa_source = QADataSource.from_db(self.db)
        qa_source.save_qa_data(charm, [])
        qa = get_missing_qa_for_queue(self.getRequest())
        self.assertEqual(1, len(qa))

    def test_charm_with_qa_data_not_in_queue(self):
        _, charm_json = factory.makeCharm(self.db, promulgated=True)
        charm = Charm(charm_json)
        qa_source = QADataSource.from_db(self.db)
        qa_source.save_qa_data(charm, [])
        qa = get_missing_qa_for_queue(self.getRequest())
        self.assertEqual(0, len(qa))

    def test_unpromulgated_with_missing_qa_not_in_queue(self):
        factory.makeCharm(self.db, promulgated=False)
        qa = get_missing_qa_for_queue(self.getRequest())
        self.assertEqual(0, len(qa))

    def test_askubuntu_questions_in_review(self):
        entry = factory.make_review_entry()
        entry['type'] = 'question'
        self.db.answer_queue.insert([entry])
        resp = charm_review(self.getRequest())
        q = resp['queue']
        self.assertEqual(1, len(q))
        self.assertEqual('question', q[0]['type'])

    def test_proofs_in_queue_with_errors(self):
        proof_with_errors = {'e': ['error 1']}
        proof_with_warnings = {'w': ['warning 1']}
        # The view retrieves promulgated charms with proof errors.
        ignore, charm_1 = factory.makeCharm(
            self.db, owner='joe', promulgated=True, proof=proof_with_errors)
        # The view ignores promulgated charms with proof warning.
        factory.makeCharm(self.db, promulgated=True, proof=proof_with_warnings)
        # The view ignored unpromulgated charms with proof errors.
        factory.makeCharm(self.db, promulgated=False, proof=proof_with_errors)
        p_queue = get_proof_errors_for_queue(self.getRequest())
        self.assertEqual(1, len(p_queue))
        self.assertEqual(
            "Proof error on %s" % charm_1['name'], p_queue[0]['summary'])

    def test_obsolete_charms_ignored_for_proofs_in_queue(self):
        proof_with_errors = {'e': ['error 1']}
        # The view retrieves promulgated charms with proof errors.
        ignore, charm_1 = factory.makeCharm(
            self.db, owner='joe', promulgated=True, proof=proof_with_errors,
            series='oneiric')
        p_queue = get_proof_errors_for_queue(self.getRequest())
        self.assertEqual([], p_queue)

    def test_proof_errors_compatible_with_review_queue(self):
        proof = {'e': ['error 1']}
        # The view retrieves promulgated charms with proof errors.
        ignore, charm = factory.makeCharm(
            self.db, owner='joe', promulgated=True, proof=proof)
        item = get_proof_errors_for_queue(self.getRequest())[0]
        self.assertEqual("Proof error on %s" % charm['name'], item['summary'])
        self.assertEqual("%s#proof" % charm['short_url'], item['item'])

    def test_store_errors_in_queue(self):
        ignore, charm = factory.makeCharm(self.db, charm_error=True)
        factory.makeCharm(self.db, charm_error=False)
        item = get_store_errors_for_queue(self.getRequest())[0]
        self.assertEqual("Store error on %s" % charm['name'], item['summary'])
        self.assertEqual(
            "http://code.launchpad.net/%s" % charm['branch_spec'],
            item['item'])

    def test_obsolete_charms_store_errors_ignored(self):
        ignore, charm = factory.makeCharm(self.db, charm_error=True,
                                            series='oneiric')
        results = get_store_errors_for_queue(self.getRequest())
        self.assertEqual([], results)


class CharmStoreMissingTestCase(ViewTestBase):

    def test_charms_with_store_errors(self):
        ignore, charm_1 = factory.makeCharm(self.db, charm_error=True)
        factory.makeCharm(self.db, charm_error=False)
        response = charm_store_missing(self.getRequest())
        charm = Charm(charm_1)
        self.assertEqual([charm], response['queue'])

    def test_obsolete_charms_ignored(self):
        ignore, charm_1 = factory.makeCharm(self.db, charm_error=True,
                                            series='oneiric')
        response = charm_store_missing(self.getRequest())
        self.assertEqual([], response['queue'])


class ProofErrorsTestCase(ViewTestBase):

    def test_promulgated_charms_with_errors(self):
        proof_with_errors = {'e': ['error 1']}
        proof_with_warnings = {'w': ['warning 1']}
        # The view retrieves promulgated charms with proof errors.
        ignore, charm_1 = factory.makeCharm(
            self.db, owner='joe', promulgated=True, proof=proof_with_errors)
        # The view ignores promulgated charms with proof warning.
        factory.makeCharm(self.db, promulgated=True, proof=proof_with_warnings)
        # The view ignored unpromulgated charms with proof errors.
        factory.makeCharm(self.db, promulgated=False, proof=proof_with_errors)
        response = proof_errors(self.getRequest())
        charm = Charm(charm_1)
        self.assertEqual([charm], response['charms'])

    def test_obsolete_charms_ignored(self):
        proof_with_errors = {'e': ['error 1']}
        # The view retrieves promulgated charms with proof errors.
        ignore, charm_1 = factory.makeCharm(
            self.db, owner='joe', promulgated=True, proof=proof_with_errors,
            series='oneiric')
        response = proof_errors(self.getRequest())
        self.assertEqual([], response['charms'])
