# Copyright 2012, 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import datetime
import itertools

import pymongo
from pyramid.view import view_config

from charmworld import cached_view_config, utils
from charmworld.models import (
    CharmSource,
    QADataSource,
)
from charmworld.views.helpers import format_proof


# 3m
REPORT_CACHE = 180


def get_sparklines(stats):
    mins = []
    maxes = []
    means = []
    for s in stats:
        mins.append(str(s['min']))
        means.append(str(s['mean']))
        maxes.append(str(s['max']))

    return {
        'mins': ','.join(mins),
        'means': ','.join(means),
        'maxes': ','.join(maxes)}


def get_missing_qa_for_queue(request):
    source = CharmSource.from_request(request)
    qa_source = QADataSource.from_db(request.db)
    qaed_charms = [(a['name'], a['series']) for a in
                   qa_source.answers.find()]
    unqaed_charms = [charm for charm in source.find_promulgated_charms() if
                     (charm.name, charm.series) not in qaed_charms]
    qa_queue = []
    for charm in unqaed_charms:
        entry = {
            'type': 'qa',
            'date_modified': 'N/A',
            'date_created': 'N/A',
            'summary': 'Missing QA data on %s' % charm.name,
            'item': charm.short_url + "/qa/edit",
            'status': 'N/A'
        }
        qa_queue.append(entry)
    return qa_queue


def get_proof_errors_for_queue(request):
    source = CharmSource.from_request(request)
    p_queue = []
    for charm in source.find_proof_error_charms(False):
        entry = {
            'type': 'proof',
            'date_modified': 'N/A',
            'date_created': 'N/A',
            'summary': 'Proof error on %s' % charm.name,
            'item': charm.short_url + "#proof",
            'status': 'N/A',
        }
        p_queue.append(entry)
    return p_queue


def get_store_errors_for_queue(request):
    source = CharmSource.from_request(request)
    s_queue = []
    for charm in source.find_mia_charms():
        entry = {
            'type': 'store',
            'date_modified': 'N/A',
            'date_created': 'N/A',
            'summary': 'Store error on %s' % charm.name,
            'item': 'http://code.launchpad.net/%s' % charm.branch_spec,
            'status': 'N/A',
        }
        s_queue.append(entry)
    return s_queue


@cached_view_config(
    route_name="tools",
    renderer="charmworld:templates/tools.pt")
def tools(request):
    return {}


@view_config(
    route_name="tools-review",
    http_cache=REPORT_CACHE,
    renderer="charmworld:templates/review.pt")
def charm_review(request):
    # This view's data in a separate cron job than ingest
    queue = list(
        request.db.review_queue.find(
            sort=[('date_modified', pymongo.ASCENDING)]))
    stats = list(request.db.review_queue_latency.find(
        date={"$gte": datetime.datetime.utcnow() - datetime.timedelta(180)},
        sort=[('date_modified', pymongo.ASCENDING)]))
    questions = list(
        request.db.answer_queue.find(
            sort=[('date_modified', pymongo.ASCENDING)]))
    proof = get_proof_errors_for_queue(request)
    store = get_store_errors_for_queue(request)
    qa = get_missing_qa_for_queue(request)
    sparklines = get_sparklines(stats)
    if stats:
        current_stats = stats[0]
        max_latency = utils.prettify_timedelta(
            datetime.timedelta(seconds=current_stats['max']))
        min_latency = utils.prettify_timedelta(
            datetime.timedelta(seconds=current_stats['min']))
        mean_latency = utils.prettify_timedelta(
            datetime.timedelta(seconds=current_stats['mean']))
    else:
        max_latency = min_latency = mean_latency = 'N/A'
    return {
        'queue': list(itertools.chain(queue, proof, store, qa, questions)),
        'age_format': utils.pretty_timedelta,
        'max': max_latency,
        'mean': mean_latency,
        'min': min_latency,
        'sparklines': sparklines}


# TODO After the store errors in review queue are approved as the way to deal
# with reviewing charm store errors, this view can go away.
@view_config(
    route_name="tools-store-missing",
    http_cache=REPORT_CACHE,
    renderer="charmworld:templates/store-missing.pt")
def charm_store_missing(request):
    queue = list(CharmSource.from_request(request).find_mia_charms())
    return {'queue': queue, 'age_format': utils.pretty_timedelta}


# TODO After the proofs in review queue are approved as the way to deal with
# reviewing charm proof errors, this view can go away.
@view_config(
    route_name="tools-proof-errors",
    http_cache=REPORT_CACHE,
    renderer="charmworld:templates/charm-proof-errors.pt")
def proof_errors(request):
    sub_only = bool(request.GET.get('sub'))
    source = CharmSource.from_request(request)
    charms = list(source.find_proof_error_charms(sub_only))

    return {
        'charms': charms,
        'format_proof': format_proof,
        'sub_only': sub_only,
    }
